/*
File Name: datatype.c
Last Modification Date:	%G%	%U%
Current Version: %M%	%I%
File Creation Date: Tue May  4 16:13:55 1999
Author: Hyeokho Choi  <choi@ece.rice.edu>

Copyright: All software, documentation, and related files in this distribution
           are Copyright (c) 1999  Rice University

Permission is granted for use and non-profit distribution providing that this
notice be clearly maintained. The right to distribute any portion for profit
or as part of any commercial product is specifically reserved for the author.

Change History:

*/

#include <malloc.h>
#include <stdio.h>
#include "hmt.h"

void nerror(error_text)
char error_text[];
 
{
  void exit();
  
  fprintf(stderr,"Numerical Recipes run-time error...\n");
  fprintf(stderr,"%s\n",error_text);
  fprintf(stderr,"...now exiting to system...\n");
  exit(1);
}

mu **mumatrix(nr,nc)
int nr,nc;
{
  int i;
  mu **m;
  m = (mu **) malloc((unsigned) nr*sizeof(mu*));
  if (!m) nerror("allocation failure 1 in mumatrix()");
  m-=1;
  
  for(i=1;i<=nr;i++) {
   m[i] = (mu*) malloc((unsigned) nc*sizeof(mu));
   if (!m[i]) nerror("allocation failure 2 in mumatrix()");
   m[i]-=1;
  }
  return m;
}

void free_mumatrix(m,nr,nc)
mu **m;
int nr,nc;
 
{
   int i;
 
   for(i=nr;i>=1;i--) free((void*) (m[i]+1));
   free((void*) (m+1));
}

si **simatrix(nr,nc)
int nr,nc;
{
  int i;
  si **m;
  m = (si **) malloc((unsigned) nr*sizeof(si*));
  if (!m) nerror("allocation failure 1 in simatrix()");
  m-=1;
  
  for(i=1;i<=nr;i++) {
   m[i] = (si*) malloc((unsigned) nc*sizeof(si));
   if (!m[i]) nerror("allocation failure 2 in simatrix()");
   m[i]-=1;
  }
  return m;
}

void free_simatrix(m,nr,nc)
si **m;
int nr,nc;
 
{
   int i;
 
   for(i=nr;i>=1;i--) free((void*) (m[i]+1));
   free((void*) (m+1));
}


ps **psmatrix(nr,nc)
int nr,nc;
{
  int i;
  ps **m;
  m = (ps **) malloc((unsigned) nr*sizeof(ps*));
  if (!m) nerror("allocation failure 1 in psmatrix()");
  m-=1;
  
  for(i=1;i<=nr;i++) {
   m[i] = (ps*) malloc((unsigned) nc*sizeof(ps));
   if (!m[i]) nerror("allocation failure 2 in psmatrix()");
   m[i]-=1;
  }
  return m;
}

void free_psmatrix(m,nr,nc)
ps **m;
int nr,nc;
 
{
   int i;
 
   for(i=nr;i>=1;i--) free((void*) (m[i]+1));
   free((void*) (m+1));
}


es **esmatrix(nr,nc)
int nr,nc;
{
  int i;
  es **m;
  m = (es **) malloc((unsigned) nr*sizeof(es*));
  if (!m) nerror("allocation failure 1 in esmatrix()");
  m-=1;
  
  for(i=1;i<=nr;i++) {
   m[i] = (es*) malloc((unsigned) nc*sizeof(es));
   if (!m[i]) nerror("allocation failure 2 in esmatrix()");
   m[i]-=1;
  }
  return m;
}

void free_esmatrix(m,nr,nc)
es **m;
int nr,nc;
 
{
   int i;
 
   for(i=nr;i>=1;i--) free((void*) (m[i]+1));
   free((void*) (m+1));
}

double *vector(nl,nh)
int nl,nh;

{
  double *v;

  v=(double*)malloc((unsigned)(nh-nl+1)*sizeof(double));
  if (!v) nerror("allocation failure in vector()");
  return v-nl;
}

int *ivector(nl,nh)
int nl,nh;

{
  int *v;

  v=(int*)malloc((unsigned)(nh-nl+1)*sizeof(int));
  if (!v) nerror("allocation failure in ivector()");
  return v-nl;
}

void free_vector(v,nl,nh)
double *v;
int nl,nh;

{
  free((char *) (v+nl));
}

void free_ivector(v,nl,nh)
int *v,nl,nh;

{
  free((char *) (v+nl));
}

double **matrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;

{
   int i;
   double **m;

   m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
   if (!m) nerror("allocation failure 1 in matrix()");
   m-=nrl;

   for(i=nrl;i<=nrh;i++) {
     m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
     if (!m[i]) nerror("allocation failure 2 in matrix()");
     m[i]-=ncl;
   }
   return m;
}

void free_matrix(m,nrl,nrh,ncl,nch)
double **m;
int nrl,nrh,ncl,nch;

{
   int i;

   for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
   free((char*) (m+nrl));
 }

