#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "hmt.h"

mu **MU,**MUO;
si **SI,**SIO;
ps **PS,**PSO;
es **ES,**ESO;
int P;
double **w;
double temp;

void main(argc,argv)
int argc;
char **argv;
{
 int i,j,level;
 double tol,nvar;
 ps **PP;

 if(argc < 3){
  printf("Usage: %s noisy_image output_image [tolerance] [size]\n",argv[0]);
  printf(" noisy_image : input image file to be denoised\n");
  printf(" output_image: denoised image output filename\n");
  printf(" [tolerance] : convergence error criterion for HMT training\n");
  printf("               Default = 1e-3. If tolerance < 0, uHMT is used.\n");
  printf(" [size]      : size of image (default = 512)\n");
  exit(0);
 }
 if (argc >=4)
  tol = atof(argv[3]);
 else
   tol = 1e-3;
 if (argc >=5)
   P = atoi(argv[4]);
 else
   P = 512;

 level = l2(P);

 MU = mumatrix(level,3);
 SI = simatrix(level,3);
 PS = psmatrix(level,3);
 ES = esmatrix(level,3);
 MUO = mumatrix(level,3);
 SIO = simatrix(level,3);
 PSO = psmatrix(level,3);
 ESO = esmatrix(level,3);
 w = matrix(1,P,1,P); 
 
 /* read in wavelet coefficient matrix */
 printf("Reading image from %s ...\n",argv[1]);
 rdmatrix(w,argv[1],P);
 dwt2daub8(w,P,1);

 printf("Estimated noise variance = %e\n",(nvar=noisevar()));


 if (tol < 0.0)
   {
     printf("Building universal HMT model ...\n");
     uhmt(MU,SI,PS,ES);
   }
 else
   {
    /* initialize the MU, SI, PS, and ES matrices */
    initialize();

    /* save old parameters */
    copy_para();
 
    printf("Training started ...\n");
    do 
      {
       /* EM algorithm */
       emhht();
       emhlt();
       emlht();
 
       /* check convergence */
       printf("%e\n",(temp=conerr()));
       copy_para();
      }
    while(temp > tol);
 
    printf("Training completed...\n");
   }

 /* compute posterior state probabilities */
 PP = psmatrix(P,P);


 if (tol < 0.0)
   {
     for(i=1;i<=level;i++)
       for(j=1;j<=3;j++)
         {
           SI[i][j].S+=nvar;
           SI[i][j].L+=nvar;
         }
     posthh(PP);
     posthl(PP);
     postlh(PP);
     emwiener(PP,nvar);
   }
 else
   {
     posthh(PP);
     posthl(PP);
     postlh(PP); 
     emwiener(PP,nvar);
   } 
 dwt2daub8(w,P,-1);
 printf("filtering completed ...\n");
 printf("Writing output image to %s ...\n",argv[2]);
 wtmatrix(w,argv[2],P);
 
}



