#include <stdio.h>
#include <math.h>
#include "hmt.h"

double *conv(double *x, int xl, double *y, int yl)
     /* xl <= yl */
{
 double *z,*temp;
 double c;
 int i,j;

 z = vector(1,xl+yl-1);
 temp = vector(1,yl+2*xl-2);

 for(i=1;i<=yl+2*xl-2;i++)
   temp[i]=0.0;
 
 for(i=1;i<=yl;i++)
   temp[i+xl-1]=y[i]; 

 for(i=1;i<=xl+yl-1;i++)
   {
     c=0.0;
     for(j=1;j<=xl;j++)
       c+=x[xl-j+1]*temp[i+j-1];
     z[i]=c;
   }
 free_vector(temp,1,yl+2*xl-2);
 return(z);
}

double *cconv(double *x, int xl, double *y, int yl)
     /* xl <= yl */
{
  double *temp, *ytemp, *z;
  int i;
  
  ytemp = vector(1,yl+xl-1);
  for(i=1;i<=yl;i++)
    ytemp[i]=y[i];

  for(i=1;i<=xl-1;i++)
    ytemp[yl+i] = ytemp[i];

  temp = vector(1,yl+2*xl-2);
  temp = conv(x,xl,ytemp,yl+xl-1);

  z = vector(1,yl);
  
  for(i=1;i<=yl;i++)
    z[i] = temp[i+xl-1];  

  free_vector(ytemp,1,yl+xl-1);
  free_vector(temp,1,yl+2*xl-2);
  return(z);
}

double *cconv1(double *x, int xl, double *y, int yl)
     /* xl <= yl */
{
  double *temp, *ytemp, *z;
  int i;
  
  ytemp = vector(1,yl+xl-1);
  for(i=1;i<=yl;i++)
    ytemp[i+xl-1]=y[i];

  for(i=xl-1;i>=1;i--)
    ytemp[i] = ytemp[i+yl];

  temp = vector(1,yl+2*xl-2);
  temp = conv(x,xl,ytemp,yl+xl-1);

  z = vector(1,yl);
  
  for(i=1;i<=yl;i++)
    z[i] = temp[i+xl-1];  

  free_vector(ytemp,1,yl+xl-1);
  free_vector(temp,1,yl+2*xl-2);
  return(z);
}

void cdwt2(double **x, int n, double *h, int hn)
{
 double *vec, *vlo, *vhi, *h0,*h1,*htemp;

 int i,j,k;

 vec = vector(1,n);
 vlo = vector(1,n);
 vhi = vector(1,n);
 h0 = vector(1,hn);
 h1 = vector(1,hn);
 htemp = vector(1,hn);

 /* build h1 here */
 for(i=1;i<=hn;i++)
   {
    h0[i]=h[hn-i+1];
    h1[i]=h[hn-i+1];
   }
 for(i=2;i<=hn;i+=2) h1[i] = -1.0*h1[i];
 for(i=1;i<=hn;i++) htemp[i]=h1[hn-i+1];
 for(i=1;i<=hn;i++) h1[i]=htemp[i];

 for(i=1;i<=n;i++)
   {
     for(j=1;j<=n;j++) vec[j]=x[i][j];
     vlo = cconv(h0,hn,vec,n);
     vhi = cconv(h1,hn,vec,n);
     for(j=1;j<=n/2;j++)
       {
         x[i][j] = vlo[2*j-1];
         x[i][n/2+j] = vhi[2*j-1];
       }
   }

   for(i=1;i<=n;i++)
   {
     for(j=1;j<=n;j++) vec[j]=x[j][i];
     vlo = cconv(h0,hn,vec,n);
     vhi = cconv(h1,hn,vec,n);
     for(j=1;j<=n/2;j++)
       {
         x[j][i] = vlo[2*j-1];
         x[n/2+j][i] = vhi[2*j-1];
       }
   }
   free_vector(vec,1,n);
   free_vector(vlo,1,n);
   free_vector(vhi,1,n);
   free_vector(h0,1,hn);
   free_vector(h1,1,hn);
   free_vector(htemp,1,hn);

   if (n/2 != 1) cdwt2(x,n/2,h,hn);
}

void cidwt2(double **x, int n, double *h, int hn)
{

 double *vecl, *vech, *vlo, *vhi, *h1;

 int i,j,k;

 if (n/2 > 1) cidwt2(x,n/2,h,hn);
 
 vecl = vector(1,n);
 vech = vector(1,n);
 vlo = vector(1,n);
 vhi = vector(1,n);
 h1 = vector(1,hn);

 /* build h1 here */
 for(i=1;i<=hn;i++) h1[i]=h[hn-i+1];
 for(i=2;i<=hn;i+=2) h1[i] = -1.0*h1[i];
 
 for(j=1;j<=n;j++) 
   {

     for(i=1;i<=n;i++)
       {
         vlo[i] = 0.0;
         vhi[i] = 0.0;
       } 
     for(i=1;i<=n/2;i++)
       {
         vlo[2*i-1] = x[i][j];
         vhi[2*i-1] = x[n/2+i][j];
       }

      vecl = cconv1(h,hn,vlo,n);
      vech = cconv1(h1,hn,vhi,n);
      for(i=1;i<=n;i++)
        x[i][j]= vecl[i]+vech[i];
   }
 
 for(j=1;j<=n;j++) 
   {

     for(i=1;i<=n;i++)
       {
         vlo[i] = 0.0;
         vhi[i] = 0.0;
       } 
     for(i=1;i<=n/2;i++)
       {
         vlo[2*i-1] = x[j][i];
         vhi[2*i-1] = x[j][n/2+i];
       }

      vecl = cconv1(h,hn,vlo,n);
      vech = cconv1(h1,hn,vhi,n);
      for(i=1;i<=n;i++)
        x[j][i]= vecl[i]+vech[i];
   }
  free_vector(vecl,1,n);
  free_vector(vech,1,n);
  free_vector(vlo,1,n);
  free_vector(vhi,1,n);
  free_vector(h1,1,hn);

}

void dwt2daub8(double **x, int n, int flag)
{
  double *h;
  int hn;

  hn = 8; /* filter length */
  h = vector(1,hn);

  /* filter coefficients */
  h[1] = 0.23037781330890;
  h[2] = 0.71484657055291;
  h[3] = 0.63088076792986;         
  h[4] = -0.02798376941686;
  h[5] = -0.18703481171909;
  h[6] = 0.03084138183556;
  h[7] = 0.03288301166689;
  h[8] = -0.01059740178507;

  if (flag == -1)
   cidwt2(x,n,h,8);
  else
   cdwt2(x,n,h,8);
  free_vector(h,1,8);
}






