/*
File Name: post.c
Last Modification Date:	%G%	%U%
Current Version: %M%	%I%
File Creation Date: Wed May  5 11:40:48 1999
Author: Hyeokho Choi  <choi@ece.rice.edu>

Copyright: All software, documentation, and related files in this distribution
           are Copyright (c) 1999  Rice University

Permission is granted for use and non-profit distribution providing that this
notice be clearly maintained. The right to distribute any portion for profit
or as part of any commercial product is specifically reserved for the author.

Change History:

*/

#include "hmt.h"
#define LOOPi for(i=1;i<=P;i++)
#define LOOPj for(j=1;j<=P;j++)
#define LOOPsi for(i=si;i<=ei;i++)
#define LOOPsj for(j=sj;j<=ej;j++)

#include "exvar.h"

void posthh(ps **P1)
{

 int i,j,k,J,J2,ip,jp,level;
 double temp,t1,t2;
 int si,ei,sj,ej,sni,snj,eni,enj;
 mu **BE, **BEP, **BER, **AL;
 mu **gtmp;
 double **scale;

 BE = mumatrix(P,P);
 BEP = mumatrix(P,P);
 BER = mumatrix(P,P);
 AL = mumatrix(P,P);

 gtmp = mumatrix(P,P);
 scale = matrix(1,P,1,P);
 level = l2(P);

 /* UP Step */
 for(k=1;k<=level;k++)
   {
     J=power2(k-1);
     J2=J*J;
     si = J+1;
     ei = 2*J;
     sj = J+1;
     ej = 2*J;
     LOOPsi
      LOOPsj
       {     
        gtmp[i][j].S = gauss(w[i][j],MU[k][1].S,SI[k][1].S);
        gtmp[i][j].L = gauss(w[i][j],MU[k][1].L,SI[k][1].L);
        scale[i][j] = (gtmp[i][j].S+gtmp[i][j].L)*0.5;
       }
   }

 si = power2(level-1)+1;
 ei = P;
 sj = power2(level-1)+1;        
 ej = P;

 LOOPsi
  LOOPsj
     {
       BE[i][j].S = gtmp[i][j].S / scale[i][j];
       BE[i][j].L = gtmp[i][j].L / scale[i][j];
     }


 for(k=level;k>=2;k--)
   {
     J=power2(k-1);
     J2=J*J;
     si = J+1;
     ei = 2*J;
     sj = J+1;
     ej = 2*J;
  
     LOOPsi
      LOOPsj
       {
         BEP[i][j].S = ES[k][1].SS*BE[i][j].S+ES[k][1].LS*BE[i][j].L;
         BEP[i][j].L = ES[k][1].SL*BE[i][j].S+ES[k][1].LL*BE[i][j].L;     
       }
      
     sni = J/2+1;
     eni = si-1;
     snj = J/2+1;
     enj = sj-1;

     /* construct Betachild matrix */
     for(i=sni;i<=eni;i++)
       for(j=snj;j<=enj;j++)
         {
           t1=BEP[2*i][2*j].S*BEP[2*i-1][2*j].S*BEP[2*i][2*j-1].S*BEP[2*i-1][2*j-1].S;
           t2=BEP[2*i][2*j].L*BEP[2*i-1][2*j].L*BEP[2*i][2*j-1].L*BEP[2*i-1][2*j-1].L;
           temp = (t1+t2)*0.5;
           scale[i][j]*=temp;
           BE[i][j].S = t1*gtmp[i][j].S/scale[i][j];
           BE[i][j].L = t2*gtmp[i][j].L/scale[i][j];
	 }

     /* construct BER matrix */
     LOOPsi
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         BER[i][j].S = BE[ip][jp].S/BEP[i][j].S;
         BER[i][j].L = BE[ip][jp].L/BEP[i][j].L;
       }
   }

 free_mumatrix(gtmp,P,P);

 /* DOWN step */
 AL[2][2].S = PS[1][1].S;
 AL[2][2].L = PS[1][1].L;

 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = J+1;
     ei = 2*J;
     sj = J+1;
     ej = 2*J;

     LOOPsi 
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         AL[i][j].S=ES[k][1].SS*AL[ip][jp].S*BER[i][j].S+ES[k][1].SL*AL[ip][jp].L*BER[i][j].L;
         AL[i][j].L=ES[k][1].LS*AL[ip][jp].S*BER[i][j].S+ES[k][1].LL*AL[ip][jp].L*BER[i][j].L;
       }
   }

 free_mumatrix(BEP,P,P);

     /* Compute Probablities */
 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = J+1;
     ei = 2*J;
     sj = J+1;
     ej = 2*J;

     LOOPsi
      LOOPsj
       {
         temp = AL[i][j].S*BE[i][j].S+AL[i][j].L*BE[i][j].L;
         P1[i][j].S= AL[i][j].S*BE[i][j].S/temp;
         P1[i][j].L= AL[i][j].L*BE[i][j].L/temp;
       }
 
     temp=AL[2][2].S*BE[2][2].S+AL[2][2].L*BE[2][2].L;
     P1[2][2].S = AL[2][2].S*BE[2][2].S/temp;
     P1[2][2].L = AL[2][2].L*BE[2][2].L/temp; 
   }
 free_mumatrix(BER,P,P);
 free_mumatrix(BE,P,P);
 free_mumatrix(AL,P,P);

}

void posthl(ps **P1)
{

 int i,j,k,J,J2,ip,jp,level;
 double temp,t1,t2;
 int si,ei,sj,ej,sni,snj,eni,enj;
 mu **BE, **BEP, **BER, **AL;
 mu **gtmp;
 double **scale;

 BE = mumatrix(P,P);
 BEP = mumatrix(P,P);
 BER = mumatrix(P,P);
 AL = mumatrix(P,P);

 gtmp = mumatrix(P,P);
 scale = matrix(1,P,1,P);
 level = l2(P);

 /* UP Step */
 for(k=1;k<=level;k++)
   {
     J=power2(k-1);
     J2 = J*J;
     si = J+1;
     ei = 2*J;
     sj = 1;
     ej = J;

     LOOPsi
      LOOPsj
       {
         gtmp[i][j].S = gauss(w[i][j],MU[k][3].S,SI[k][3].S);
         gtmp[i][j].L = gauss(w[i][j],MU[k][3].L,SI[k][3].L);
         scale[i][j] = (gtmp[i][j].S+gtmp[i][j].L)*0.5;
       }
   }

 si = power2(level-1)+1;
 ei = P;
 sj = 1;        
 ej = power2(level-1);

 LOOPsi
  LOOPsj
     {
       BE[i][j].S = gtmp[i][j].S / scale[i][j];
       BE[i][j].L = gtmp[i][j].L / scale[i][j];
     }


 for(k=level;k>=2;k--)
   {
     J=power2(k-1);
     J2=J*J;
     si = J+1;
     ei = 2*J;
     sj = 1;
     ej = J;
  
     LOOPsi
      LOOPsj
       {
         BEP[i][j].S = ES[k][3].SS*BE[i][j].S+ES[k][3].LS*BE[i][j].L;
         BEP[i][j].L = ES[k][3].SL*BE[i][j].S+ES[k][3].LL*BE[i][j].L;     
       }
      
     sni = J/2+1;
     eni = si-1;
     snj = 1;
     enj = J/2;

     /* construct Betachild matrix */
     for(i=sni;i<=eni;i++)
       for(j=snj;j<=enj;j++)
         {
           t1=BEP[2*i][2*j].S*BEP[2*i-1][2*j].S*BEP[2*i][2*j-1].S*BEP[2*i-1][2*j-1].S;
           t2=BEP[2*i][2*j].L*BEP[2*i-1][2*j].L*BEP[2*i][2*j-1].L*BEP[2*i-1][2*j-1].L;
           temp = (t1+t2)*0.5;
           scale[i][j]*=temp;
           BE[i][j].S = t1*gtmp[i][j].S/scale[i][j];
           BE[i][j].L = t2*gtmp[i][j].L/scale[i][j];
	 }

     /* construct BER matrix */
     LOOPsi
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         BER[i][j].S = BE[ip][jp].S/BEP[i][j].S;
         BER[i][j].L = BE[ip][jp].L/BEP[i][j].L;
       }
   }

 free_mumatrix(gtmp,P,P);

 /* DOWN step */
 AL[2][1].S = PS[1][3].S;
 AL[2][1].L = PS[1][3].L;

 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = J+1;
     ei = 2*J;
     sj = 1;
     ej = J;

     LOOPsi 
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         AL[i][j].S=ES[k][3].SS*AL[ip][jp].S*BER[i][j].S+ES[k][3].SL*AL[ip][jp].L*BER[i][j].L;
         AL[i][j].L=ES[k][3].LS*AL[ip][jp].S*BER[i][j].S+ES[k][3].LL*AL[ip][jp].L*BER[i][j].L;
       }
   }

 free_mumatrix(BEP,P,P);

     /* Compute Probablities */
 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = J+1;
     ei = 2*J;
     sj = 1;
     ej = J;

     LOOPsi
      LOOPsj
       {
         temp = AL[i][j].S*BE[i][j].S+AL[i][j].L*BE[i][j].L;
         P1[i][j].S= AL[i][j].S*BE[i][j].S/temp;
         P1[i][j].L= AL[i][j].L*BE[i][j].L/temp;
       }
 
     temp=AL[2][1].S*BE[2][1].S+AL[2][1].L*BE[2][1].L;
     P1[2][1].S = AL[2][1].S*BE[2][1].S/temp;
     P1[2][1].L = AL[2][1].L*BE[2][1].L/temp; 
   }
 free_mumatrix(BER,P,P);
 free_mumatrix(BE,P,P);
 free_mumatrix(AL,P,P);
}

void postlh(ps **P1)
{

 int i,j,k,J,J2,ip,jp,level;
 double temp,t1,t2;
 int si,ei,sj,ej,sni,snj,eni,enj;
 mu **BE, **BEP, **BER, **AL;
 mu **gtmp;
 double **scale;

 BE = mumatrix(P,P);
 BEP = mumatrix(P,P);
 BER = mumatrix(P,P);
 AL = mumatrix(P,P);

 gtmp = mumatrix(P,P);
 scale = matrix(1,P,1,P);
 level = l2(P);

 /* UP Step */
 for(k=1;k<=level;k++)
   {
     J=power2(k-1);
     J2 = J*J;
     si = 1;
     ei = J;
     sj = J+1;
     ej = 2*J;

     LOOPsi
      LOOPsj
       {
        gtmp[i][j].S = gauss(w[i][j],MU[k][2].S,SI[k][2].S);
        gtmp[i][j].L = gauss(w[i][j],MU[k][2].L,SI[k][2].L);
        scale[i][j] = (gtmp[i][j].S+gtmp[i][j].L)*0.5;
       } 
  }

 si = 1;
 ei = power2(level-1);
 sj = power2(level-1)+1;        
 ej = P;

 LOOPsi
  LOOPsj
     {
       BE[i][j].S = gtmp[i][j].S / scale[i][j];
       BE[i][j].L = gtmp[i][j].L / scale[i][j];
     }


 for(k=level;k>=2;k--)
   {
     J=power2(k-1);
     J2=J*J;
     si = 1;
     ei = J;
     sj = J+1;
     ej = 2*J;
  
     LOOPsi
      LOOPsj
       {
         BEP[i][j].S = ES[k][2].SS*BE[i][j].S+ES[k][2].LS*BE[i][j].L;
         BEP[i][j].L = ES[k][2].SL*BE[i][j].S+ES[k][2].LL*BE[i][j].L;     
       }
      
     sni = 1;
     eni = J/2;
     snj = J/2+1;
     enj = sj-1;

     /* construct Betachild matrix */
     for(i=sni;i<=eni;i++)
       for(j=snj;j<=enj;j++)
         {
           t1=BEP[2*i][2*j].S*BEP[2*i-1][2*j].S*BEP[2*i][2*j-1].S*BEP[2*i-1][2*j-1].S;
           t2=BEP[2*i][2*j].L*BEP[2*i-1][2*j].L*BEP[2*i][2*j-1].L*BEP[2*i-1][2*j-1].L;
           temp = (t1+t2)*0.5;
           scale[i][j]*=temp;
           BE[i][j].S = t1*gtmp[i][j].S/scale[i][j];
           BE[i][j].L = t2*gtmp[i][j].L/scale[i][j];
	 }

     /* construct BER matrix */
     LOOPsi
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         BER[i][j].S = BE[ip][jp].S/BEP[i][j].S;
         BER[i][j].L = BE[ip][jp].L/BEP[i][j].L;
       }
   }


 free_mumatrix(gtmp,P,P);

 /* DOWN step */
 AL[1][2].S = PS[1][2].S;
 AL[1][2].L = PS[1][2].L;

 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = 1;
     ei = J;
     sj = J+1;
     ej = 2*J;

     LOOPsi 
      LOOPsj
       {
         ip = (i+1)/2;
         jp = (j+1)/2;
         AL[i][j].S=ES[k][2].SS*AL[ip][jp].S*BER[i][j].S+ES[k][2].SL*AL[ip][jp].L*BER[i][j].L;
         AL[i][j].L=ES[k][2].LS*AL[ip][jp].S*BER[i][j].S+ES[k][2].LL*AL[ip][jp].L*BER[i][j].L;
       }
   }

 free_mumatrix(BEP,P,P);

     /* Compute Probablities */
 for(k=2;k<=level;k++)
   {
     J = power2(k-1);
     J2 = J*J;
     si = 1;
     ei = J;
     sj = J+1;
     ej = 2*J;

     LOOPsi
      LOOPsj
       {
         temp = AL[i][j].S*BE[i][j].S+AL[i][j].L*BE[i][j].L;
         P1[i][j].S= AL[i][j].S*BE[i][j].S/temp;
         P1[i][j].L= AL[i][j].L*BE[i][j].L/temp;
       }
 
     temp=AL[1][2].S*BE[1][2].S+AL[1][2].L*BE[1][2].L;
     P1[1][2].S = AL[1][2].S*BE[1][2].S/temp;
     P1[1][2].L = AL[1][2].L*BE[1][2].L/temp; 
   }

 free_mumatrix(BER,P,P);
 free_mumatrix(BE,P,P);
 free_mumatrix(AL,P,P);
}






