#include "hmt.h"
#include <math.h>
#include "exvar.h"
#define PI 3.14159265358979323846 

int l2(int p)
{
 return( (int) (log((double) p)/log(2.0)+0.5));
}

double gauss(double w, double m ,double s)
{
  double y;

  y = 1.0/sqrt(2.0*PI*s)*exp(-(w-m)*(w-m)*0.5/s)+1e-15;
  return(y);
}

double conerr()
{
 int i,j,J,level;
 double errval;
 double t1,t2,t3,t4;
 double pserr,muerr,sierr,eserr;

 pserr = 0.0;
 muerr = 0.0;
 sierr = 0.0;
 eserr = 0.0;
 level = l2(P);
 J = level*3;

 for(i=1;i<=level;i++)
   for(j=1;j<=3;j++)
     {
       t1 = PS[i][j].S-PSO[i][j].S;
       t2 = PS[i][j].L-PSO[i][j].L;
       pserr+=t1*t1+t2*t2;
       /* t1 = (MU[i][j].S-MUO[i][j].S);
       t2 = (MU[i][j].L-MUO[i][j].L;
       muerr+=t1*t1+t2*t2; */
       t1 = (SI[i][j].S-SIO[i][j].S)/SI[i][j].S;
       t2 = (SI[i][j].L-SIO[i][j].L)/SI[i][j].L;
       sierr+=t1*t1+t2*t2;
       t1 = ES[i][j].SS-ESO[i][j].SS;
       t2 = ES[i][j].SL-ESO[i][j].SL;
       t3 = ES[i][j].LS-ESO[i][j].LS;
       t4 = ES[i][j].LL-ESO[i][j].LL;
       eserr+=t1*t1+t2*t2+t3*t3+t4*t4;
     }

 pserr/=J;
 /* muerr/=J; */
 sierr/=J;
 eserr/=J;

 errval = pserr;
 /* if (muerr > errval) errval = muerr; */ 
 if (sierr > errval) errval = sierr;
 if (eserr > errval) errval = eserr;
 return(errval);
}

int power2(i)
int i;
{
 int j,k;
 k=1;
 for(j=1;j<=i;j++) k*=2;
 return(k);
}   
             
void copy_para()
{
 int i,j,level;
 
 level = l2(P);
 for(i=1;i<=level;i++)
  for(j=1;j<=3;j++)
    {
      MUO[i][j].S = MU[i][j].S;
      MUO[i][j].L = MU[i][j].L;
      SIO[i][j].S = SI[i][j].S;
      SIO[i][j].L = SI[i][j].L;
      PSO[i][j].S = PS[i][j].S;
      PSO[i][j].L = PS[i][j].L;
      ESO[i][j].SS = ES[i][j].SS;
      ESO[i][j].SL = ES[i][j].SL;
      ESO[i][j].LS = ES[i][j].LS;
      ESO[i][j].LL = ES[i][j].LL;
    }
} 

void initialize()
{
 int i,j,k,level;
 int J,J2,si,ei,sj,ej;
 double tsis,tsil,temp;

 level = l2(P);

 for(k=1;k<=level;k++)
   {
     J=power2(k-1);
     J2=J*J;
     
     /* HH subband */
     si=J+1;
     ei=2*J;
     sj=J+1;
     ej=2*J;
     tsis = 0.0;
     tsil = 0.0;
     
     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
         { 
           temp=w[i][j];
           tsis+=temp*temp;
           temp=w[i][j];
           tsil+=temp*temp;
	 }

     tsis/=J2;
     tsil/=J2;

     MU[k][1].S = 0.0;
     MU[k][1].L = 0.0;
     ES[k][1].SS=0.5;
     ES[k][1].SL=0.5;
     ES[k][1].LS=0.5;
     ES[k][1].LL=0.5;
     PS[k][1].S = 0.5;
     PS[k][1].L = 0.5;
     SI[k][1].S = tsis*0.5;
     SI[k][1].L = tsil*2.0;
                    
     /* LH subband */
     si=1;
     ei=J;
     sj=J+1;
     ej=2*J;
     tsis = 0.0;
     tsil = 0.0;

     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
         { 
           temp=w[i][j];
           tsis+=temp*temp;
           temp=w[i][j];
           tsil+=temp*temp;
	 }

     tsis/=J2;
     tsil/=J2;
     
     MU[k][2].S = 0.0;
     MU[k][2].L = 0.0;
     ES[k][2].SS=0.5;
     ES[k][2].SL=0.5;
     ES[k][2].LS=0.5;
     ES[k][2].LL=0.5;
     PS[k][2].S = 0.5;
     PS[k][2].L = 0.5;
     SI[k][2].S = tsis*0.5;
     SI[k][2].L = tsil*2.0;

     /* HL subband */
     si=J+1;
     ei=2*J;
     sj=1;
     ej=J;
     tsis = 0.0;
     tsil = 0.0;

     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
         { 
           temp=w[i][j];
           tsis+=temp*temp;
           temp=w[i][j];
           tsil+=temp*temp;
	 }

     tsis/=J2;
     tsil/=J2;

     MU[k][3].S = 0.0;
     MU[k][3].L = 0.0;
     ES[k][3].SS=0.5;
     ES[k][3].SL=0.5;
     ES[k][3].LS=0.5;
     ES[k][3].LL=0.5;
     PS[k][3].S = 0.5;
     PS[k][3].L = 0.5;
     SI[k][3].S = tsis*0.5;
     SI[k][3].L = tsil*2.0;     
   }
}

double noisevar()
{
 int i,j;
 double x;
 
 x = 0.0;
 for(i=P/2+1;i<=P;i++)
   for(j=P/2+1;j<=P;j++)
     x+= w[i][j]*w[i][j];
 x/=0.25*P*P;
 return(x);
}

void emwiener(ps **PP,double nv)
{
 int i,j,k,J,J2,si,ei,sj,ej,level,slev;
 double y,sps,spl;
 
 slev =5;
 level = l2(P);
 for(k=slev;k<=level;k++)
   {
     J=power2(k-1);
     J2=J*J;
     
     /* HH subband */
     si=J+1;
     ei=2*J;
     sj=J+1;
     ej=2*J;

     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
	{
         sps = SI[k][1].S - nv;
         if (sps < 0.0) sps = 0.0;
         spl = SI[k][1].L - nv;
         if (spl < 0.0) spl = 0.0;
         y = w[i][j]*(sps/(sps+nv)*PP[i][j].S+spl/(spl+nv)*PP[i][j].L);     
         w[i][j] = y;
	}

      /* LH subband */
     si=1;
     ei=J;
     sj=J+1;
     ej=2*J;

     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
	{
         sps = SI[k][2].S - nv;
         if (sps < 0.0) sps = 0.0;
         spl = SI[k][2].L - nv;
         if (spl < 0.0) spl = 0.0;
         y = w[i][j]*(sps/(sps+nv)*PP[i][j].S+spl/(spl+nv)*PP[i][j].L);     
         w[i][j] = y;
	}

      /* HL subband */
     si=J+1;
     ei=2*J;
     sj=1;
     ej=J;

     for(i=si;i<=ei;i++)
       for(j=sj;j<=ej;j++)
	{
         sps = SI[k][3].S - nv;
         if (sps < 0.0) sps = 0.0;
         spl = SI[k][3].L - nv;
         if (spl < 0.0) spl = 0.0;
         y = w[i][j]*(sps/(sps+nv)*PP[i][j].S+spl/(spl+nv)*PP[i][j].L);     
         w[i][j] = y;
	}  
   }
}



