/*
File Name: uhmt.c
Last Modification Date:	%G%	%U%
Current Version: %M%	%I%
File Creation Date: Mon May 10 12:49:14 1999
Author: Hyeokho Choi  <choi@ece.rice.edu>

Copyright: All software, documentation, and related files in this distribution
           are Copyright (c) 1999  Rice University

Permission is granted for use and non-profit distribution providing that this
notice be clearly maintained. The right to distribute any portion for profit
or as part of any commercial product is specifically reserved for the author.

Change History:

*/

#include <math.h>
#include "hmt.h"
#include "exvar.h"

void uhmt()
{
  int lev;
  double alpha_big, c1_big, alpha_sm, c1_sm, beta;
  int i,j;

  lev = l2(P);
  alpha_big = 2.5;
  c1_big = 13.0;
  alpha_sm = 2.5;
  c1_sm = 7.0;
  beta = 1.0;

  /* variances decay exponentially */

  for(j=1;j<=3;j++)
    {
      PS[1][j].S = 0.5;
      PS[1][j].L = 0.5;

      for(i=1;i<=lev;i++)
        {
          SI[i][j].S = pow(2.0,c1_sm)*pow(2.0,-alpha_sm*i);
          SI[i][j].L = pow(2.0,c1_big)*pow(2.0,-alpha_big*i);
        }
      for(i=1;i<=3;i++)
        {
          ES[i][j].SS = 1.0;
          ES[i][j].LL = 1.0;
	}
      for(i=4;i<=lev;i++)
        {
          ES[i][j].SS = 0.8+0.2*(1.0-pow(2.0,-1.0*(beta*(i-4.0))));
          ES[i][j].LL = 0.9-0.4*(1.0-pow(2.0,-1.0*(beta*(i-4.0))));
        }
      for(i=1;i<=lev;i++)
        {
          ES[i][j].LS = 1.0 - ES[i][j].SS;
          ES[i][j].SL = 1.0 - ES[i][j].LL;
          MU[i][j].S = 0.0;
          MU[i][j].L = 0.0;
        }
      for(i=2;i<=lev;i++)
        {
          PS[i][j].S = PS[i-1][j].S*ES[i][j].SS+PS[i-1][j].L*ES[i][j].SL;
          PS[i][j].L = PS[i-1][j].S*ES[i][j].LS+PS[i-1][j].L*ES[i][j].LL;
        }
    }
}

  
